import re
from typing import Any, List, Optional, Tuple, Union

primitive_type = (int, float, str, bool)
uuid_pattern = re.compile(r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$")


def is_primitive_type(value: Any) -> bool:
    return isinstance(value, primitive_type)


def is_integer(value: str) -> bool:
    return to_integer(value) is not None


def to_integer(value: str) -> Optional[int]:
    if isinstance(value, str) and (value := value.strip()):
        try:
            return int(value)
        except Exception:
            pass
    return None


def is_float(value: str) -> bool:
    return to_float(value) is not None


def to_float(value: str) -> Optional[int]:
    if isinstance(value, str) and (value := value.strip()):
        try:
            return float(value)
        except Exception:
            pass
    return None


def any_of_bool(*booleans) -> bool:
    for boolean in booleans:
        if boolean is True:
            return True
    return False


def at_most_one_of_bool(*booleans) -> bool:
    ntrue = 0
    for boolean in booleans:
        if boolean is True:
            if ntrue > 0:
                return False
            ntrue += 1
    return True


def to_bool(value: str) -> bool:
    return value if isinstance(value, bool) else ((value.strip().lower() == "true")
                                                  if isinstance(value, str) else False)


def to_string_list(value: Union[List[str], Tuple[str, ...], str], strip: bool = True, empty: bool = True) -> List[str]:
    strings = []
    if isinstance(value, (list, tuple)):
        for item in value:
            if isinstance(item, str):
                if (strip is not False):
                    item = item.strip()
                if (empty is True) or item:
                    strings.append(item)
    elif isinstance(value, str):
        if (strip is not False):
            value = value.strip()
        if (empty is True) or value:
            strings.append(value)
    return strings


def to_non_empty_string_list(value: Union[List[str], Tuple[str, ...], str], strip: bool = True) -> List[str]:
    return to_string_list(value, strip=strip, empty=False)


def is_uuid(value: str) -> bool:
    return uuid_pattern.match(value) if isinstance(value, str) else False


def to_flattened_list(*args) -> List[Any]:
    flattened_list = []
    def flatten(arg):  # noqa
        if isinstance(arg, (list, tuple, set)):
            for item in arg:
                flatten(item)
        else:
            flattened_list.append(arg)
    flatten(args)
    return flattened_list



import iterm2
import json


def tohtml(text):
    return "<pre>" + text.replace("&", "&amp;").replace("<", "&lt;") + "</pre>"


def prettyprint(text):
    try:
        root = json.loads(text)
    except json.decoder.JSONDecodeError as e:
        return "Invalid JSON: {}".format(e)
    return json.dumps(root, sort_keys=True, indent=4)


async def main(connection):
    app = await iterm2.async_get_app(connection)

    # Set the click handler
    @iterm2.RPC
    async def onclick(session_id):
        session = app.get_session_by_id(session_id)
        selection = await session.async_get_selection()
        selectedText = await session.async_get_selection_text(selection)
        await component.async_open_popover(session_id, tohtml(prettyprint(selectedText)), iterm2.util.Size(600, 600))

    # Define the configuration knobs:
    vl = "json_pretty_printer"
    knobs = [iterm2.CheckboxKnob("JSON Pretty Printer", False, vl)]
    component = iterm2.StatusBarComponent(
        short_description="JSON Pretty Printer",
        detailed_description=("Select JSON in the terminal,"
                              " then click this status bar component to see it nicely formatted."),
        knobs=knobs,
        exemplar="JSON",
        update_cadence=None,
        identifier="com.iterm2.json-pretty-printer")

    # This function gets called whenever any of the paths named in defaults (below) changes
    # or its configuration changes.
    @iterm2.StatusBarRPC
    async def coro(knobs):
        return ["JSON"]

    # Register the component.
    await component.async_register(connection, coro, onclick=onclick)


iterm2.run_forever(main)



import os
import requests
from typing import Optional


def get_github_commit_date(repo: str, commit: str, github_token: Optional[str] = None,
                           raise_exception: bool = False) -> Optional[str]:
    try:
        request = f"{repo}/commits/{commit}"
        if (response := _get_github_info(request, github_token=github_token, raise_exception=raise_exception)):
            return response.get("commit", {}).get("author", {}).get("date")
    except Exception as e:
        if raise_exception is True:
            raise e
    return None


def get_github_latest_commit(repo: str, branch: str, github_token: Optional[str] = None,
                             raise_exception: bool = False) -> Optional[str]:
    try:
        request = f"{repo}/branches/{branch}"
        if (response := _get_github_info(request, github_token=github_token, raise_exception=raise_exception)):
            return response.get("commit", {}).get("sha")
    except Exception as e:
        if raise_exception is True:
            raise e
    return None


def _get_github_info(request: str, github_token: Optional[str] = None, raise_exception: bool = False) -> Optional[dict]:
    url = f"https://api.github.com/repos/{request}"
    if (not github_token) and (not (github_token := os.environ.get("GITHUB_TOKEN"))):
        return None
    try:
        headers = {"Authorization": f"token {github_token}"}
        if (response := requests.get(url, headers=headers)).status_code == 200:
            return response.json()
    except Exception as e:
        if raise_exception is True:
            raise e
    return None



from __future__ import annotations
import os
import re
import sys
from typing import Any, Callable, List, Optional, Tuple, Union
from hms_utils.chars import chars
from hms_utils.dictionary_utils import load_json_file
from hms_utils.dictionary_parented import DictionaryParented as JSON
from hms_utils.path_utils import basename_path, repack_path, unpack_path
from hms_utils.type_utils import is_primitive_type


class ConfigBasic:

    _PATH_SEPARATOR = "/"
    _PATH_COMPONENT_CURRENT = "."
    _PATH_COMPONENT_PARENT = ".."
    _PATH_COMPONENT_ROOT = _PATH_SEPARATOR
    _MACRO_START = "${"
    _MACRO_END = "}"
    _MACRO_PATTERN = re.compile(r"\$\{([^}]+)\}")
    _MACRO_HIDE_START = "@@@@@@@__macro_hide_start_["
    _MACRO_HIDE_END = "]__macro_hide_end_@@@@@@@"
    _MACRO_PATTERN = re.compile(r"\$\{([^}]+)\}")
    _EXPORT_NAME_SEPARATOR = ":"
    _TRICKY_FIX = True

    def __init__(self, config: Union[dict, str],
                 name: Optional[str] = None,
                 path_separator: Optional[str] = None,
                 decrypted: bool = False,
                 custom_macro_lookup: Optional[Callable] = None,
                 raise_exception: bool = False) -> None:

        if not (isinstance(path_separator, str) and (path_separator := path_separator.strip())):
            path_separator = ConfigBasic._PATH_SEPARATOR
        if isinstance(config, str):
            config = load_json_file(config)
        elif isinstance(config, JSON) or (not isinstance(config, dict)):
            raise Exception("Must create Config object with dictionary or file path.")
        self._json = self._create_json(config)
        self._name = name if isinstance(name, str) and name else None
        self._includes = None
        self._path_separator = path_separator
        self._custom_macro_lookup = custom_macro_lookup if callable(custom_macro_lookup) else None
        self._decrypted = decrypted is True
        self._ignore_missing_macros = True
        self._ignore_circular_macros = True
        self._ignore_structured_macros = True
        self._warnings = []
        self._raise_exception = raise_exception is True
        self._merged = []

    def _create_json(self, data: dict) -> JSON:
        return JSON(data)

    def data(self) -> JSON:
        return self._json.sorted()

    @property
    def path_separator(self) -> str:
        return self._path_separator

    @property
    def decrypted(self) -> bool:
        return self._decrypted

    @property
    def json(self) -> JSON:
        return self._json

    @property
    def name(self) -> Optional[str]:
        return self._name

    def merge(self, data: Union[Union[dict, ConfigBasic],
                                List[Union[dict, ConfigBasic]]]) -> Tuple[List[str], List[str]]:
        merged_paths = [] ; unmerged_paths = []  # noqa
        if isinstance(data, (dict, ConfigBasic)):
            data = [data]
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict) and (not isinstance(item, JSON)):
                    item = JSON(item)
                elif isinstance(item, ConfigBasic):
                    if merged_name := item.name:
                        self._merged.append(merged_name)
                    item = item._json
                if isinstance(item, JSON):
                    self._json, item_merged_paths, item_unmerged_paths = (
                        self._json.merge(item, path_separator=self._path_separator))
                    merged_paths.extend(item_merged_paths)
                    unmerged_paths.extend(item_unmerged_paths)
        return merged_paths, unmerged_paths

    def include(self, data: Union[List[Union[dict, ConfigBasic]], Union[dict, ConfigBasic]]) -> None:
        if isinstance(data, (dict, ConfigBasic)):
            data = [data]
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict) and (not isinstance(item, JSON)):
                    item = ConfigBasic(item)
                if isinstance(item, ConfigBasic):
                    if self._includes is None:
                        self._includes = []
                    self._includes.append(item)
                    if item.secrets:
                        self._secrets = True

    def lookup(self, path: str,
               context: Optional[JSON] = None,
               noexpand: bool = False,
               inherit_simple: bool = False,
               inherit_none: bool = False, **kwargs) -> Optional[Union[Any, JSON]]:
        if isinstance(context, JSON):
            specified_context = context
        else:
            context = self._json
            specified_context = None
        value, context = self._lookup(path, context=context, inherit_simple=inherit_simple, inherit_none=inherit_none)
        if (value is not None) and (noexpand is not True):
            # N.B. context_path only used for warning message from _note_macro_not_found.
            value = self.expand_macros(value, context=specified_context or context, context_path=path)
        if (value is None) and self._includes:
            for includes in self._includes:
                if (value := includes.lookup(path, noexpand=noexpand,
                                             inherit_simple=inherit_simple,
                                             inherit_none=inherit_none, **kwargs)) is not None:
                    break
        if path.endswith(self.path_separator) and isinstance(value, JSON):
            value = value.duplicate()
            # TODO: Referencing kwargs here (for show from ConfigWithSecrets.lookup) is slightly icky.
            if inherited_values := self.lookup_inherited_values(value, **kwargs):
                for inherited_value_key in inherited_values:
                    if inherited_value_key not in value:
                        value[inherited_value_key] = inherited_values[inherited_value_key]
        return value

    def lookup_inherited_values(self, value: JSON, **kwargs) -> dict:
        results = {}
        context = value
        if isinstance(value, JSON):
            parent = value.parent
            while parent:
                for key in parent:
                    if not isinstance(parent[key], dict):
                        if (not isinstance(value, dict) or (key not in value)) and (key not in results):
                            path = self.path(parent, path_suffix=key)
                            if value := self.lookup(path, context=context, **kwargs):
                                if is_primitive_type(value):
                                    results[key] = value
                parent = parent.parent
        return results

    def exports(self, lookup_paths: List[str], show: Optional[bool] = False) -> Tuple[dict, int]:
        make_export_key = lambda key: basename_path(key).replace("-", "_")  # noqa
        exports = {} ; status = 0  # noqa
        if isinstance(lookup_paths, str):
            lookup_paths = [lookup_paths]
        if not (isinstance(lookup_paths, list) and lookup_paths):
            return exports, status
        for lookup_path in lookup_paths:
            if (index := lookup_path.find(ConfigBasic._EXPORT_NAME_SEPARATOR)) > 0:
                exports_name = lookup_path[0:index]
                if not (lookup_path := lookup_path[index + 1:].strip()):
                    continue
            else:
                exports_name = basename_path(lookup_path)
            if (value := self.lookup(lookup_path, show=show)) is None:
                status = 1
                continue
            # Since dash is not even allowed in environment/export name change to underscore.
            if isinstance(value, JSON):
                for key in value:
                    key_value = value[key]
                    if is_primitive_type(key_value):
                        exports[make_export_key(key)] = key_value
            else:
                exports[make_export_key(exports_name)] = value
            if isinstance(value, JSON) and (inherited_values := self.lookup_inherited_values(value, show=show)):
                for inherited_value_key in inherited_values:
                    if (export_key := make_export_key(inherited_value_key)) not in exports:
                        exports[export_key] = inherited_values[inherited_value_key]
        exports = dict(sorted(exports.items()))
        for export_key in exports:
            if ConfigBasic._contains_macro(exports[export_key]):
                status = 1
                break
        return exports, status

    def _lookup(self, path: str, context: Optional[JSON] = None,
                inherit_simple: bool = False, inherit_none: bool = False) -> Tuple[Optional[Union[Any, JSON]], JSON]:

        def lookup_path_components(path_components: List[str], context: JSON) -> Tuple[Optional[Any], JSON]:
            value = None
            for path_component_index, path_component in enumerate(path_components):
                if (value := context.get(path_component)) is None:
                    break
                if isinstance(value, JSON):
                    # Found a JSON in the path so recurse down to it.
                    context = value
                elif path_component_index < (len(path_components) - 1):
                    # Found a terminal (non-JSON) in the path but it is not the last component.
                    value = None
                    break
            return value, context, path_component_index

        secondary_contexts = []
        if context is None:
            context = self._json
        elif isinstance(context, list) and context:
            secondary_contexts = context[1:]
            context = context[0]
        elif not isinstance(context, JSON):
            context = self._json

        value = None

        if not (path_components := self.unpack_path(path)):
            return value, context
        if path_rooted := (path_components[0] == ConfigBasic._PATH_COMPONENT_ROOT):
            if not (path_components := path_components[1:]):
                return context, context
            context = context.root

        value, context, path_component_index = lookup_path_components(path_components, context)
        if (value is None) and secondary_contexts:
            for secondary_context in secondary_contexts:
                value, _, _ = lookup_path_components(path_components, secondary_context)

        if (value is None) and (inherit_none is not True) and context.parent:
            #
            # Search for the remaining path up through parents simulating inheritance.
            # Disable this behavior with the inherit_none flag. And if inherit_simple
            # is set then we only do this behavior for the last component of a path.
            # For example if we have this hierarchy:
            #
            #   - portal:
            #     - identity: identity_value
            #     - auth:
            #       - client: auth_client_value
            #     - smaht:
            #       - wolf:
            #         - some_property: some_property_value
            #
            # When we lookup /portal/smaht/wolf/identity we would get the value for /portal/identity,
            # identity_value, because it is visible within the /portal/smaht/wolf context; and if
            # we lookup /portal/wolf/auth/client we would get the value for /portal/auth/client,
            # auth_client_value, because it - /portal/auth - is also visible with that context.
            # Now if the inherit_simple flag is set then the first case gets the same result, but the
            # latter case - /portal/smaht/wolf/auth/client will return None because we would be looking
            # for a path with more than one component - auth/client - within the inherited/parent contexts.
            # The inherit_simple case is in case it turns out the that non-simple case is not very intuitive.
            #
            path_components_left = path_components[0:max(0, path_component_index - 1)]
            path_components_right = path_components[path_component_index:]
            if (inherit_simple is not True) or len(path_components_right) == 1:
                # This is a bit tricky; and note we lookup in parent but return current context.
                path_components = path_components_left + path_components_right
                path = self.repack_path(path_components, path_rooted=path_rooted)
                lookup_value, lookup_context = self._lookup(path, context=context.parent)
                if ConfigBasic._TRICKY_FIX and (lookup_value is not None):
                    context = ([context, *lookup_context]
                               if isinstance(lookup_context, list) else [context, lookup_context])
                return lookup_value, context

        return value, context

    def expand_macros(self, value: Any, context: Optional[JSON] = None, context_path: Optional[str] = None) -> Any:
        # Note FYI that we do not macros the nested of macros.
        if isinstance(value, str):
            value = self._expand_macros_within_string(value, context, context_path=context_path)
        elif isinstance(value, JSON):
            for key in value:
                value[key] = self.expand_macros(value[key], context=value, context_path=context_path)
        return value

    def _expand_macros_within_string(self, value: str,
                                     context: Optional[JSON] = None, context_path: Optional[str] = None) -> Any:

        def hide_macros(value: str, macro_values: Union[str, List[str]]) -> str:
            for macro_value in macro_values if isinstance(macro_values, (list, set)) else [macro_values]:
                value = value.replace(f"{ConfigBasic._MACRO_START}{macro_value}{ConfigBasic._MACRO_END}",
                                      f"{ConfigBasic._MACRO_HIDE_START}{macro_value}{ConfigBasic._MACRO_HIDE_END}")
            return value

        def unhide_macros(value: str) -> str:
            return (value.replace(ConfigBasic._MACRO_HIDE_START, ConfigBasic._MACRO_START)
                         .replace(ConfigBasic._MACRO_HIDE_END, ConfigBasic._MACRO_END))

        if not (isinstance(value, str) and value):
            return value

        expanding_macros = []
        missing_macro_found = False
        resolved_macro_context = None

        while True:
            if not ((match := ConfigBasic._MACRO_PATTERN.search(value)) and (macro_value := match.group(1))):
                break
            # This is a bit tricky with the context.
            resolved_macro_value, resolved_macro_context = self.lookup_macro(macro_value,
                                                                             context=resolved_macro_context or context)
            if resolved_macro_value is not None:
                if not is_primitive_type(resolved_macro_value):
                    self._warning(f"Macro must resolve to primitive type:"
                                  f" {self.path(context, path_suffix=macro_value)}", not self._ignore_structured_macros)
                    return value
                value = value.replace(f"${{{macro_value}}}", str(resolved_macro_value))
                if macro_value in expanding_macros:
                    self._warning(f"Circular macro definition found: {macro_value}", not self._ignore_circular_macros)
                    value = hide_macros(value, expanding_macros)
                    missing_macro_found = True
                else:
                    expanding_macros.append(macro_value)
            else:
                self._note_macro_not_found(macro_value, context, context_path=context_path)
                value = hide_macros(value, macro_value)
                missing_macro_found = True
        if missing_macro_found:
            value = unhide_macros(value)
        return value

    def lookup_macro(self, macro_value: str, context: Optional[JSON] = None) -> Tuple[Any, JSON]:
        if self.is_absolute_path(macro_value):
            resolved_macro_value = self.lookup(macro_value)
            resolved_macro_context = context
        else:
            resolved_macro_value, resolved_macro_context = self._lookup(macro_value, context=context)
            if (resolved_macro_value is None) and self._custom_macro_lookup:
                resolved_macro_value = self._custom_macro_lookup(macro_value, resolved_macro_context)
        if (resolved_macro_value is None) and self._includes:
            for include_item in self._includes:
                resolved_macro_value, resolved_macro_context = include_item.lookup_macro(macro_value)
                if resolved_macro_value is not None:
                    break
        return resolved_macro_value, resolved_macro_context

    def unpack_path(self, path: str) -> List[str]:
        return unpack_path(path, path_separator=self._path_separator,
                           path_current=ConfigBasic._PATH_COMPONENT_CURRENT,
                           path_parent=ConfigBasic._PATH_COMPONENT_PARENT)

    def repack_path(self, path_components: List[str], path_rooted: bool = False) -> str:
        return repack_path(path_components, path_rooted=path_rooted, path_separator=self._path_separator)

    def normalize_path(self, path: str) -> str:
        return self.repack_path(self.unpack_path(path))

    def is_absolute_path(self, path: str) -> bool:
        return isinstance(path, str) and path.startswith(self._path_separator)

    def context_path(self, context: JSON, path_suffix: Optional[str] = None) -> str:
        return context.context_path(path_rooted=True, path_suffix=path_suffix)

    def path(self, context: JSON, path_rooted: bool = True, path_suffix: Optional[str] = None) -> str:
        if not isinstance(context, JSON):
            return ""
        return context.context_path(path_separator=self._path_separator,
                                    path_rooted=path_rooted, path_suffix=path_suffix)

    def _contains_macro(value: Any) -> bool:
        def contains_macro(value: str) -> bool:
            return isinstance(value, str) and ConfigBasic._MACRO_PATTERN.search(value) is not None
        if is_primitive_type(value):
            return contains_macro(value)
        elif isinstance(value, dict):
            for key in value:
                if contains_macro(value[key]):
                    return True
        elif isinstance(value, list):
            for element in value:
                if contains_macro(value[key]):
                    return True
        return False

    def _note_macro_not_found(self, macro_value: str,
                              context: Optional[JSON] = JSON, context_path: Optional[str] = None) -> None:
        if isinstance(context, list) and context:
            # TODO: one of tests ends up here with context as a list; why.
            # Oh yeah, I'm using a list of contexts in some cases.
            context = context[0]
        if not (isinstance(context_path, str) and context_path):
            if isinstance(context, JSON):
                context_path = context.context_path(path_separator=self._path_separator, path_rooted=True)
            else:
                context_path = None
        self._warning(f"Macro not found: {macro_value}"
                      f"{f' {chars.dot} context: {context_path}' if context_path else ''}",
                      not self._ignore_missing_macros)

    def _warning(self, message: str, raise_exception: bool = False) -> None:
        if (raise_exception is True) or self._raise_exception:
            raise Exception(message)
        if isinstance(message, str) and message:
            message = f"WARNING: {message}"
        if message not in self._warnings:
            self._warnings.append(message)

    def _debug(self, message: Union[str, Callable]) -> None:
        if ("--debug" in sys.argv) or ("-debug" in sys.argv) or (os.environ.get("HMS_DEBUG", "").lower() == "true"):
            if callable(message):
                message = message()
            if isinstance(message, str):
                print("DEBUG: " + message, file=sys.stderr, flush=True)



import base64
from typing import Optional
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import os


def encrypt_file(plaintext_file: str, password: str, encrypted_file: str, prefix: Optional[str] = None) -> bool:
    if not (isinstance(prefix, str) and (prefix := prefix.strip())):
        prefix = b"__HMS_CRYPTO__"
    with open(plaintext_file, "rb") as f:
        data = f.read()
    encrypted_data = Fernet(_derive_key_from_password(password, salt := os.urandom(16))).encrypt(data)
    encrypted_data = prefix + b"\n" + base64.urlsafe_b64encode(salt) + b"\n" + base64.urlsafe_b64encode(encrypted_data)
    with open(f"{encrypted_file}", "wb") as f:
        f.write(encrypted_data)
    return True


def decrypt_file(encrypted_file: str, password: str, decrypted_file: str, prefix: Optional[str] = None) -> bool:
    decrypted_data = read_encrypted_file(encrypted_file, password, prefix=prefix)
    with open(decrypted_file, "wb") as f:
        f.write(decrypted_data)
    return True


def read_encrypted_file(encrypted_file: str, password: str, prefix: Optional[str] = None) -> str:
    with open(encrypted_file, "rb") as f:
        data = f.read().split(b"\n")
    salt, encrypted_data = base64.urlsafe_b64decode(data[1]), base64.urlsafe_b64decode(data[2])
    return Fernet(_derive_key_from_password(password, salt)).decrypt(encrypted_data)


def _derive_key_from_password(password: str, salt: bytes) -> bytes:
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000, backend=default_backend())
    return base64.urlsafe_b64encode(kdf.derive(password.encode()))



from prettytable import PrettyTable, HRuleStyle
from typing import List
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.portal.portal_utils import Portal as Portal


def main():

    argv = ARGV({
        ARGV.OPTIONAL(str): ["--env"],
        ARGV.OPTIONAL(str): ["query"],
        ARGV.OPTIONAL(bool): ["--current"],
        ARGV.OPTIONAL(bool): ["--deleted"],
        ARGV.OPTIONAL(bool): ["--inactive"],
        ARGV.OPTIONAL(bool): ["--revoked"],
        ARGV.OPTIONAL(bool): ["--submitters", "--submitter"],
        ARGV.OPTIONAL(bool): ["--non-submitters", "--non-submitter",
                              "--nonsubmitters", "--nonsubmitter", "--nosubmitters", "--nosubmitter"],
        ARGV.OPTIONAL(bool): ["--admin"],
        ARGV.OPTIONAL(bool): ["--database"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(bool): ["--ping"],
        ARGV.OPTIONAL(str): ["--submission-center", "--submission-centers", "--centers", "--center", "--sc"]
    })

    portal = Portal.create(argv.env, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)

    if argv.query:
        argv.query = argv.query.lower()
    if argv.current:
        status_query = "status=current"
    elif argv.deleted:
        status_query = "status=deleted&status=revoked"
    elif argv.inactive:
        status_query = "status=inactive"
    elif argv.revoked:
        status_query = "status=revoked"
    else:
        status_query = "status=current&status=inactive&status=deleted&status=revoked"
    if not (users := portal.get_metadata(f"/users?{status_query}", limit=10000, database=argv.database)):
        return
    users = users.get("@graph")

    table = PrettyTable()
    table.field_names = ["N", "USER", "NAME", "CONSORTIA", "CENTERS", "GROUP", "STATUS"]
    table.align = "l"
    table.align["N"] = "r"
    if argv.verbose:
        table.hrules = HRuleStyle.ALL

    users = sorted(users, key=lambda item: item.get("email"))

    ordinal = 1
    for user in users:
        group = _get_group_display_value(user)
        if argv.admin and ("admin" not in group):
            if "admin" not in group:
                continue
        submission_centers = _get_submission_centers_display_value(user, verbose=argv.verbose)
        if argv.submission_center:
            if (argv.submission_center.strip().lower() and
                (argv.submission_center.strip().lower() not in ["none", "null"])):  # noqa
                if (argv.submission_center.strip().lower() not in submission_centers.lower()):
                    continue
            if (argv.submission_center.strip().lower() in ["none", "null"]) and submission_centers:
                continue
        user_uuid = user.get("uuid", "")
        user_email = user.get("email", "")
        user_first_name = user.get("first_name", "")
        user_last_name = user.get("last_name", "")
        user_status = user.get("status")
        user_groups = _get_group_display_value(user) or chars.null
        user_submission_centers = submission_centers or chars.null
        if (user_submits_for := _get_submits_for(user)):
            if argv.non_submitters:
                continue
            if (user_submits_for == _get_submission_centers(user)):
                user_submission_centers += f" {chars.check}"
            else:
                user_submission_centers += f" Σ"
        elif argv.submitters:
            continue
        if argv.query:
            if not ((argv.query in user_uuid.lower()) or
                    (argv.query in user_email.lower()) or
                    (argv.query in user_first_name.lower()) or
                    (argv.query in user_last_name.lower())):
                continue
        table.add_row([
            ordinal,
            user_email + ("\n" + user_uuid if argv.verbose else ""),
            user_first_name + " " + user_last_name,
            _get_consortia_display_value(user, verbose=argv.verbose) or chars.null,
            user_submission_centers,
            user_groups,
            user_status
        ])
        ordinal += 1
    if ordinal > 1:
        print(table)


def _get_consortia_display_value(user: dict, verbose: bool = False) -> str:
    values = [] ; values_verbose = ""  # noqa
    if isinstance(consortia := user.get("consortia"), list):
        for consortium in consortia:
            if isinstance(value := consortium.get("identifier"), str):
                values.append(value)
                if (verbose is True) and (value_uuid := consortium.get("uuid")):
                    values_verbose += "\n" + value_uuid
    values = ", ".join(values)
    if values_verbose:
        values += values_verbose
    return values


def _get_submission_centers_display_value(user: dict, verbose: bool = False) -> str:
    values = [] ; values_verbose = ""  # noqa
    if isinstance(submission_centers := user.get("submission_centers"), list):
        for submission_center in submission_centers:
            if isinstance(value := submission_center.get("identifier"), str):
                values.append(value)
                if (verbose is True) and (value_uuid := submission_center.get("uuid")):
                    values_verbose += "\n" + value_uuid
    values = ", ".join(values)
    if values_verbose:
        values += values_verbose
    return values


def _get_submission_centers(user: dict) -> List[str]:
    values = []
    if isinstance(submission_centers := user.get("submission_centers"), list):
        for submission_center in submission_centers:
            if isinstance(value := submission_center.get("identifier"), str):
                if value not in values:
                    values.append(value)
    return sorted(values)


def _get_submits_for_display_value(user: dict) -> str:
    values = []
    if isinstance(submits_for := user.get("submits_for"), list):
        for submit_for in submits_for:
            if isinstance(value := submit_for.get("identifier"), str):
                values.append(value)
    values = ", ".join(values)
    return values


def _get_submits_for(user: dict) -> List[str]:
    values = []
    if isinstance(submission_centers := user.get("submission_centers"), list):
        for submission_center in submission_centers:
            if isinstance(value := submission_center.get("identifier"), str):
                if value not in values:
                    values.append(value)
    return sorted(values)


def _get_group_display_value(user: dict) -> str:
    values = []
    if isinstance(user, dict):
        if isinstance(groups := user.get("groups"), list):
            values = groups
    return ", ".join(values)


if __name__ == "__main__":
    main()



from getpass import getpass as read_password
import os
import shutil
import sys
from dcicutils.command_utils import yes_or_no
from dcicutils.tmpfile_utils import create_temporary_file_name, temporary_file
from hms_utils.crypt_utils import decrypt_file, encrypt_file
from hms_utils.argv import ARGV


def main():

    STDIN = "/dev/stdin"
    STDOUT = "/dev/stdout"

    argv = ARGV({
        ARGV.OPTIONAL(bool): ["--encrypt"],
        ARGV.OPTIONAL(bool): ["--decrypt"],
        ARGV.OPTIONAL(str): ["--output", "--out"],
        ARGV.OPTIONAL(bool): ["--yes", "--force"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(str): ["--password", "--passwd"],
        ARGV.REQUIRED(str): ["file", ARGV.DEFAULT, "-"],
        ARGV.EXACTLY_ONE_OF: ["--encrypt", "--decrypt"],
    })

    def copy_file(source: str, destination: str) -> None:
        nonlocal STDOUT
        if destination == STDOUT:
            with open(source, "r") as f:
                sys.stdout.write(f.read())
        else:
            shutil.copy(source, destination)

    if argv.encrypt:
        verb = "Encrypt"
        function = encrypt_file
    elif argv.decrypt:
        verb = "Decrypt"
        function = decrypt_file

    if argv.file == "-":
        argv.file = STDIN
    if (argv.file != STDIN) and (not os.path.isfile(argv.file)):
        _error(f"Cannot find file: {argv.file}")

    if argv.output:
        if (argv.output == "-") or (argv.output == STDOUT):
            argv.output = STDOUT
        elif os.path.isfile(argv.output):
            if not (argv.yes or yes_or_no(f"Overwrite file: {argv.output} ?")):
                exit(0)
        elif (output_directory := os.path.dirname(argv.output)) and (not os.path.isdir(output_directory)):
            _error(f"Output file directory does not exist: {output_directory}")
        elif argv.file == STDIN:
            argv.output = STDOUT
    elif argv.file == STDIN:
        argv.output = STDOUT
    elif argv.yes or yes_or_no(f"{verb} file {argv.file} in place?"):
        argv.output = argv.file
    else:
        argv.output = None

    if not argv.password:
        argv.password = read_password("Enter password: ")
        if not argv.password:
            _error("Must specify a password.")
        if function == encrypt_file:
            verify_password = read_password("Verify password: ")
            if verify_password != argv.password:
                _error("Passwords to not match.")

    with temporary_file() as tmpfile:

        if argv.debug:
            _print(f"{verb}ing file {argv.file} to {tmpfile}")
        elif argv.verbose and (argv.output == argv.file):
            _print(f"{verb}ing file in place: {argv.file}")
        function(argv.file, argv.password, tmpfile)
        if argv.debug:
            _print(f"Done {verb.lower()}ing file {argv.file} to {tmpfile}")
        if argv.output:
            if argv.debug:
                _print(f"Copying file {tmpfile} to: {argv.output}")
            copy_file(tmpfile, argv.output)
            if argv.debug:
                _print(f"Done copying file {tmpfile} to: {argv.output}")
        else:
            argv.output = create_temporary_file_name()
            copy_file(tmpfile, argv.output)
            _print(f"{verb}ed file is here: {argv.output}")


def main_encrypt():
    sys.argv.insert(1, "--encrypt")
    main()


def main_decrypt():
    sys.argv.insert(1, "--decrypt")
    main()


def _print(message: str) -> None:
    print(message, file=sys.stderr)


def _error(message: str) -> None:
    print(message, file=sys.stderr)
    exit(1)


def _usage():
    _error("usage: encrypt/decrypt --password password file")
    exit(1)


if __name__ == "__main__":
    sys.argv = sys.argv
    main()



from __future__ import annotations
from collections import defaultdict, deque
from copy import deepcopy
import glob
import io
import json
import os
from typing import Any, Callable, Iterator, List, Optional, Tuple, Union
from hms_utils.type_utils import is_uuid, to_non_empty_string_list


def delete_paths_from_dictionary(data: dict, paths: List[str], separator: str = "/", copy: bool = True):
    """
    Deletes from the given dictionary the keys/properies specified in the given list of key paths;
    the paths being (by default) a hierarchical-like slash-separated key names, e.g. abc/def/ghi.
    If the copy flag is True this a copy is made of the given dictionary, otherwise it is
    changed in place. In any case returns the resultant dictionary
    """
    if copy is not False:
        data = deepcopy(data)
    def delete(data, keys):  # noqa
        if len(keys) == 1:
            data.pop(keys[0], None)
        else:
            key = keys[0]
            if key in data and isinstance(data[key], dict):
                delete(data[key], keys[1:])
            if not data[key]:
                data.pop(key, None)
    for path in paths:
        keys = path.split('/')
        delete(data, keys)
    return data


def delete_properties_from_dictionaries(data: Union[List[dict], dict], properties: List[str]) -> None:
    """
    Deletes, in place, from the given dictionary or list, recursively, and/all properties
    within dictionaries with a name which is in the list of given properties.
    """
    if properties := to_non_empty_string_list(properties):
        if isinstance(data, list):
            for element in data:
                delete_properties_from_dictionaries(element, properties)
        elif isinstance(data, dict):
            for key in list(data.keys()):
                if key in properties:
                    del data[key]
                else:
                    delete_properties_from_dictionaries(data[key], properties)


def sort_dictionary(data: dict, reverse: bool = False, sensitive: bool = False, lists: bool = False) -> dict:
    """
    Sorts the given dictionary and returns the result; does not change the given dictionary.
    """
    if isinstance(data, list):
        sorted_list = []
        for element in data:
            sorted_list.append(sort_dictionary(element))
        return sorted_list
    if not isinstance(data, dict):
        return data
    sorted_data = {}
    def sorted_key(key: Any) -> str:  # noqa
        if not isinstance(key, str):
            key = str(key)
        return key.lower() if (sensitive is True) else key
    for key in sorted(data.keys(), reverse=reverse is True, key=sorted_key):
        sorted_data[key] = sort_dictionary(data[key], reverse=reverse)
    return sorted_data


def load_json_file(file: str, raise_exception: bool = False) -> Optional[dict]:
    if isinstance(file, str) and os.path.isfile(file := os.path.expanduser(file)):
        try:
            with io.open(file, "r") as f:
                return json.load(f)
        except Exception as e:
            if raise_exception is True:
                raise e
        return None


def get_referenced_uuids(item: Union[dict, List[dict]],
                         ignore_uuids: Optional[List[str]] = None,
                         exclude_uuid: bool = False,
                         exclude_properties: Optional[List[str]] = None,
                         uuid_property_name: Optional[str] = None,
                         include_paths: bool = False) -> List[str]:
    """
    Returns a list of uuids which appear as a value anywhere, recursively, within the given
    dictionary or list. But ignores any which are in the given ignore_uuids list. If the
    given exclude_uuid flag is set then ignores values of dictionary properties whose key
    is named "uuid" (or named for the give uuid_property_name). If the given included_paths
    flag is set then also looks for uuid values as a part of slash-separated path values.
    """
    referenced_uuids = []
    def find_referenced_uuids(item: Any) -> None:  # noqa
        nonlocal referenced_uuids, ignore_uuids
        if isinstance(item, dict):
            for key in item:
                if (not isinstance(exclude_properties, list)) or (key not in exclude_properties):
                    find_referenced_uuids(item[key])
        elif isinstance(item, (list, tuple)):
            for element in item:
                find_referenced_uuids(element)
        elif uuids := get_uuids_from_value(item):
            for uuid in uuids:
                if (uuid not in ignore_uuids) and (uuid not in referenced_uuids):
                    referenced_uuids.append(uuid)
    def get_uuids_from_value(value: str) -> List[str]:  # noqa
        uuids = []
        if isinstance(value, str):
            if is_uuid(value):
                uuids.append(value)
            elif include_paths is True:
                for component in value.split("/"):
                    if is_uuid(component := component.strip()):
                        uuids.append(component)
        return uuids
    if not isinstance(ignore_uuids, list):
        ignore_uuids = []
    if not (isinstance(uuid_property_name, str) and uuid_property_name):
        uuid_property_name = "uuid"
    if isinstance(item, dict):
        if (exclude_uuid is True) and (uuid := item.get(uuid_property_name)) and (uuid not in ignore_uuids):
            ignore_uuids.append(uuid)
    elif isinstance(item, list):
        for element in item:
            if ((exclude_uuid is True) and isinstance(element, dict) and
                (uuid := element.get(uuid_property_name)) and (uuid not in ignore_uuids)):  # noqa
                ignore_uuids.append(uuid)
    find_referenced_uuids(item)
    return referenced_uuids


def get_referenced_uuids_from_file(file: str,
                                   ignore_uuids: Optional[List[str]] = None, exclude_uuid: bool = False) -> List[str]:
    try:
        with io.open(file, "r") as f:
            return get_referenced_uuids(json.load(f), ignore_uuids=ignore_uuids, exclude_uuid=exclude_uuid)
    except Exception:
        return []


def get_referenced_uuids_from_files(directory: str,
                                    ignore_uuids: Optional[List[str]] = None, exclude_uuid: bool = False) -> List[str]:
    referenced_uuids = []
    try:
        for file in glob.glob(os.path.join(directory, '*.json')):
            for uuid in get_referenced_uuids_from_file(file, ignore_uuids=ignore_uuids, exclude_uuid=exclude_uuid):
                if uuid not in referenced_uuids:
                    referenced_uuids.append(uuid)
    except Exception:
        pass
    return referenced_uuids


def get_uuids(data: Union[dict, list], uuid_property_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of uuids which appear as a values of properties with a key named "uuid" (or
    name for the given uuid_property_name), recursively, within the given dictionary or list.
    """
    uuids = []
    def get_uuids(data: Union[dict, list]) -> None:  # noqa
        nonlocal uuids, uuid_property_name
        if isinstance(data, list):
            for element in data:
                get_uuids(element)
        elif isinstance(data, dict):
            if uuid := data.get(uuid_property_name):
                if is_uuid(uuid) and (uuid not in uuids):
                    uuids.append(uuid)
            else:
                for value in data.values():
                    if isinstance(value, (dict, list)):
                        get_uuids(value)
    if not (isinstance(uuid_property_name, str) and uuid_property_name):
        uuid_property_name = "uuid"
    get_uuids(data)
    return uuids


def contains_uuid(data: Union[dict, list], uuid: str, uuid_property_name: Optional[str] = None) -> bool:
    """
    Returns true iff the given dictionary or list contains, recursively, a dictionary
    property named "uuid" (or named for the given uuid_property_name) which has a
    value which is equal to the given uuid value; otherwise returns false.
    """
    def contains_uuid(data: Union[dict, list]) -> bool:
        nonlocal uuid
        if isinstance(data, list):
            for element in data:
                if contains_uuid(element):
                    return True
        elif isinstance(data, dict):
            if data.get(uuid_property_name) == uuid:
                return True
            for value in data.values():
                if isinstance(value, (dict, list)):
                    if contains_uuid(value):
                        return True
    if not (isinstance(uuid_property_name, str) and uuid_property_name):
        uuid_property_name = "uuid"
    return contains_uuid(data) if is_uuid(uuid) else False


def find_dictionary_item(items: List[dict], property_value: Any, property_name: Optional[str] = None) -> Optional[int]:
    """
    Finds the (first) dictionary in the given list whose "uuid" property (or the property named
    by the given property_name) value matches the given property_value and returns its list index,
    if found; if not found then returns None.
    """
    if isinstance(items, list) and (property_value is not None):
        if not (isinstance(property_name, str) and property_name):
            property_name = "uuid"
        for i in range(len(items)):
            if isinstance(item := items[i], dict):
                if item.get(property_name) == property_value:
                    return i
    return None


def get_property(data: dict, name: str, fallback: Optional[Any] = None) -> Optional[Any]:
    """
    Returns the value of the given property name within the given dictionary, where the given
    property name can be a dot-separated list of property names, which indicate a path into
    nested dictionaries within the given dictionary; returns None if not found.
    """
    if isinstance(data, dict) and isinstance(name, str) and name:
        if keys := name.split("."):
            nkeys = len(keys) ; key_index_max = nkeys - 1  # noqa
            for key_index in range(nkeys):
                if (value := data.get(keys[key_index], None)) is not None:
                    if key_index == key_index_max:
                        return value
                    elif isinstance(value, dict):
                        data = value
                        continue
                break
    return fallback


def get_properties(data: dict, name: str, fallback: Optional[Any] = None, sort: bool = False) -> List[Any]:
    """
    Returns the values of the given property name within the given dictionary as a list, where the
    given property name can be a dot-separated list of property names, which indicate a path into
    nested dictionaries within the given dictionary; and - where if any of the elements within
    the path are lists then we iterate through each, collecting the values for each and including
    each within the list of returned values.
    """
    if isinstance(data, dict) and isinstance(name, str) and name:
        if keys := name.split("."):
            nkeys = len(keys) ; key_index_max = nkeys - 1  # noqa
            for key_index in range(nkeys):
                if (value := data.get(keys[key_index], None)) is not None:
                    if key_index == key_index_max:
                        return [value]
                    elif isinstance(value, dict):
                        data = value
                        continue
                    elif isinstance(value, list) and value and ((sub_key_index := key_index + 1) < nkeys):
                        sub_key = ".".join(keys[sub_key_index:])
                        values = []
                        for element in value:
                            if isinstance(element_value := get_properties(element, sub_key), list):
                                for element_value_item in element_value:
                                    if (element_value_item is not None) and (element_value_item not in values):
                                        values.append(element_value_item)
                            elif (element_value is not None) and (element_value not in values):
                                values.append(element_value)
                        return sorted(values) if (sort is True) else values
                break
    return fallback if isinstance(fallback, list) else ([] if fallback is None else [fallback])


def select_items(items: List[dict], predicate: Callable) -> List[dict]:
    if not (isinstance(items, list) and items):
        return []
    return [item for item in items if predicate(item)]


def group_items_by(items: list[dict], grouping: str,
                   sort: bool = False,
                   noitems: bool = False,
                   omit_unique_items_count: bool = False,
                   map_grouping_value: Optional[Callable] = None,
                   prefix_grouping_value: bool = False,
                   identifying_property: Optional[str] = "uuid",
                   raw: bool = False) -> dict:
    if not (isinstance(items, list) and items and isinstance(grouping, str) and (grouping := grouping.strip())):
        return {}
    if not callable(map_grouping_value):
        map_grouping_value = None
    if not (isinstance(identifying_property, str) and (identifying_property := identifying_property.strip())):
        identifying_property = None
    # Initialize results with first None element to make sure items which are not
    # part of a group are listed first; delete later of no such (ungrouped) items;
    # though if sort is True then this is irrelevant.
    non_unique_item_count = 0
    results = {None: 0 if noitems is True else []}
    for item in items:
        if identifying_property and ((identifying_value := item.get(identifying_property)) is not None):
            item_identity = identifying_value
        else:
            item_identity = item
        if grouping_values := get_properties(item, grouping):
            for grouping_value in grouping_values:
                # This prefixing with the grouping name was added later when we realized it is useful to have,
                # for each individual item grouped, the name of the grouping for which it is from.
                if map_grouping_value:
                    if (grouping_value := map_grouping_value(grouping, grouping_value)) is None:
                        continue
                if prefix_grouping_value is True:
                    grouping_value = f"{grouping}:{grouping_value}"
                if noitems is True:
                    if results.get(grouping_value) is None:
                        results[grouping_value] = 0
                    results[grouping_value] += 1
                else:
                    if results.get(grouping_value) is None:
                        results[grouping_value] = []
                    results[grouping_value].append(item_identity)
                non_unique_item_count += 1
        elif noitems is True:
            results[None] += 1
            non_unique_item_count += 1
        else:
            results[None].append(item_identity)
            non_unique_item_count += 1
    if not results[None]:
        del results[None]
    if sort is True:
        # Currently sort means to sort the groups in descending order of the
        # number of items in each group list; and secondarily by the group value.
        if noitems is True:
            results = dict(sorted(results.items(), key=lambda item: (-item[1], item[0] is None, item[0] or "")))
        else:
            results = dict(sorted(results.items(), key=lambda item: (-len(item[1]), item[0] is None, item[0] or "")))
    if (raw is True) or (not results):
        return results
    results = {
        "group": grouping,
        "item_count": non_unique_item_count,
        "unique_item_count": len(items),
        "group_count": len(results),
        "group_items": results
    }
    if omit_unique_items_count is True:
        del results["unique_item_count"]
    return results


def group_items_by_groupings(items: List[dict], groupings: List[str],
                             sort: bool = False,
                             noitems: bool = False,
                             omit_unique_items_count: bool = False,
                             map_grouping_value: Optional[Callable] = None,
                             prefix_grouping_value: bool = False,
                             identifying_property: Optional[str] = "uuid") -> dict:
    if not (isinstance(items, list) and items):
        return {}
    if isinstance(groupings, str) and groupings:
        groupings = [groupings]
    elif not (isinstance(groupings, list) and groupings):
        return {}
    if not (isinstance(identifying_property, str) and (identifying_property := identifying_property.strip())):
        identifying_property = None
    if not (grouped_items := group_items_by(items, groupings[0], sort=sort,
                                            omit_unique_items_count=omit_unique_items_count,
                                            map_grouping_value=map_grouping_value,
                                            prefix_grouping_value=prefix_grouping_value,
                                            identifying_property=identifying_property)):
        return {}
    def sub_group_items_by(group_items: dict, grouping: str) -> None:  # noqa
        nonlocal items, sort, omit_unique_items_count, map_grouping_value, identifying_property
        for grouped_item_key in group_items:
            if isinstance(group_items[grouped_item_key], list):
                sub_items = select_items(
                    items, lambda item: item.get(identifying_property) in group_items[grouped_item_key])
                group_items[grouped_item_key] = group_items_by(
                    sub_items, grouping, sort=sort,
                    omit_unique_items_count=omit_unique_items_count,
                    map_grouping_value=map_grouping_value,
                    prefix_grouping_value=prefix_grouping_value,
                    identifying_property=identifying_property)
            elif isinstance(group_items[grouped_item_key], dict):
                sub_group_items_by(group_items[grouped_item_key]["group_items"], grouping)
    for grouping in groupings[1:]:
        if isinstance(grouping, str) and (grouping := grouping.strip()):
            sub_group_items_by(grouped_items["group_items"], grouping)
    if noitems is True:
        def change_group_items_list_to_group_items_count(grouped_items: dict) -> None:
            if isinstance(group_items := grouped_items.get("group_items"), dict):
                for group_item_key in group_items:
                    if isinstance(group_items[group_item_key], dict):
                        change_group_items_list_to_group_items_count(group_items[group_item_key])
                    elif isinstance(group_items[group_item_key], list):
                        group_items[group_item_key] = len(group_items[group_item_key])
        change_group_items_list_to_group_items_count(grouped_items)
    return grouped_items


def compare_dictionaries_ordered(a: dict, b: dict) -> bool:
    # CAVEAT: Written by ChatGPT and used without almost no review (just testing)!
    def compare_lists_ordered(a: list, b: list):
        if len(a) != len(b):
            return False
        for aitem, bitem in zip(a, b):
            if isinstance(aitem, dict) and isinstance(bitem, dict):
                if not compare_dictionaries_ordered(aitem, bitem):
                    return False
            elif isinstance(aitem, list) and isinstance(bitem, list):
                if not compare_lists_ordered(aitem, bitem):
                    return False
            elif aitem != bitem:
                return False
        return True
    if not isinstance(a, dict) or not isinstance(b, dict):
        return False
    if list(a.keys()) != list(b.keys()):
        return False
    for key in a:
        avalue = a[key]
        bvalue = b[key]
        if isinstance(avalue, dict) and isinstance(bvalue, dict):
            if not compare_dictionaries_ordered(avalue, bvalue):
                return False
        elif isinstance(avalue, list) and isinstance(bvalue, list):
            if not compare_lists_ordered(avalue, bvalue):
                return False
        elif avalue != bvalue:
            return False
    return True


def order_dictionary_by_dependencies(items: List[dict],
                                     dependencies: Union[List[str], str, Callable],
                                     identifying_property_name: str = "uuid") -> List[dict]:
    """
    Orders the given list of items (dictionaries) so that each item appears after its dependencies
    as defined by the given dependencies argument, where this can be a property name or a list of
    property names of identifying values, or a callable which returns the identifying values for
    a given item. There must be a single identifying value for each record whose property name
    is specified by the given identifying_property_name argument, or is "uuid" if none.
    Returns the ordered list of items; does NOT make changes in place.
    CAVEAT: This function was adapted from ChatGPT generated code.
    """
    if not isinstance(items, list):
        return []
    if not (isinstance(identifying_property_name, str) and identifying_property_name):
        identifying_property_name = "uuid"
    if not callable(dependencies):
        dependent_property_names = []
        if isinstance(dependencies, list):
            for dependency in dependencies:
                if isinstance(dependency, str) and dependency:
                    dependent_property_names.append(dependency)
        elif isinstance(dependencies, str) and dependencies:
            dependent_property_names = [dependencies]
        if not dependent_property_names:
            return items
        def get_dependencies(item: dict) -> List[str]:  # noqa
            nonlocal dependent_property_names
            dependency_values = []
            for dependent_property_name in dependent_property_names:
                if (dependency_value := item.get(dependent_property_name)) is not None:
                    if isinstance(dependency_value_list := dependency_value, list):
                        for dependency_value in dependency_value_list:
                            if dependency_value is not None:
                                dependency_values.append(dependency_value)
                    else:
                        dependency_values.append(dependency_value)
            return dependency_values
        dependencies = get_dependencies

    # Create a map from identifying-value to item.
    identifying_value_to_item_map = {item[identifying_property_name]: item for item in items}

    # Create adjacency list and in-degree count for topological sort.
    adjacency_list = defaultdict(list)
    in_degree = defaultdict(int)

    # Build graph.
    for item in items:
        identifying_value = item[identifying_property_name]
        if isinstance(dependency_values := dependencies(item), list):
            for dependency_value in dependency_values:
                if dependency_value is not None:
                    adjacency_list[dependency_value].append(identifying_value)
                    in_degree[identifying_value] += 1

    # Topological sort.
    sorted_items = []
    zero_in_degree = deque([identifying_value for identifying_value in identifying_value_to_item_map
                            if in_degree[identifying_value] == 0])
    while zero_in_degree:
        current_identifying_value = zero_in_degree.popleft()
        sorted_items.append(identifying_value_to_item_map[current_identifying_value])
        for dependent_value in adjacency_list[current_identifying_value]:
            in_degree[dependent_value] -= 1
            if in_degree[dependent_value] == 0:
                zero_in_degree.append(dependent_value)

    # Check for cycles (unsatisfied dependencies).
    if len(sorted_items) != len(items):
        raise ValueError("The input contains cyclic dependencies and cannot be ordered.")

    return sorted_items


def normalize_elastic_search_aggregation_results(data: dict) -> dict:
    # CAVEAT: Written by ChatGPT and used with only a little review (just testing)!
    def get_first_field_with_buckets_list_property(data: dict) -> Optional[dict]:  # noqa 
        if isinstance(data, dict):
            for key in data:
                if isinstance(data[key], dict) and isinstance(data[key].get("buckets"), list):
                    return data[key]
        return None
    def process_field(field: dict) -> None:  # noqa
        if not isinstance(field, dict):
            return
        if not isinstance(buckets := field.get("buckets"), list):
            return
        group_name = field.get("meta", {}).get("field_name")
        group_items = {}
        item_count = 0
        for bucket in buckets:
            if not (key := bucket.get("key_as_string")):
                if (key := bucket.get("key")) in ["No value", "null", "None", None]:
                    key = None
            doc_count = bucket["doc_count"]
            item_count += doc_count
            if nested_field := get_first_field_with_buckets_list_property(bucket):
                group_items[key] = process_field(nested_field)
            else:
                group_items[key] = doc_count
        return {
            "group": group_name,
            "item_count": item_count,
            "group_count": len(group_items),
            "group_items": group_items,
        }
    if not isinstance(data, dict):
        return {}
    return process_field(get_first_field_with_buckets_list_property(data))


# THIS WILL GO AWAY (and using one in dicationary_parented) WHEN hms_config is obsoleted to config/cli.
# This JSON class isa dictionary type which also suport "parent" property for each/every sub-dictionary
# within the main dictionary. Should be able to use EXACTLY like a dicttionary type after creating with
# JSON(dict); including copying and setting properties to any type including either other dictionaies
# or JSON objects. Also has a "root" property on each sub-dictionary referring to the main/root one;
# this just walks up the parent properties to the top (where parent is of course None).
#
class JSON(dict):

    def __init__(self, data: Optional[Union[dict, JSON]] = None,
                 read_value: Optional[Callable] = None,
                 write_value: Optional[Callable] = None,
                 _initializing: bool = False) -> None:
        if isinstance(data, JSON):
            data = deepcopy(dict(data)) if _initializing is not True else dict(data)
        elif not isinstance(data, dict):
            data = {}
        super().__init__(data)
        self._initialized = False
        self._parent = None
        self._write_value = write_value if callable(write_value) else lambda value: value
        self._read_value = read_value if callable(read_value) else lambda value: value

    @property
    def parent(self) -> Optional[JSON]:
        return self._parent

    def items(self) -> Iterator[Tuple[Any, Any]]:
        self._initialize()
        for key, value in super().items():
            if isinstance(value, dict) and (not isinstance(value, JSON)):
                value = JSON(value)
            yield key, value

    def values(self) -> Iterator[Any]:
        self._initialize()
        return super().values()

    @property
    def root(self) -> Optional[JSON]:
        node = self
        while True:
            if node._parent is None:
                return node
            node = node._parent

    @property
    def context_path(self) -> Optional[str]:
        # FYI we only actually use this (in hms_config) for diagnostic messages.
        context = self
        context_path = []
        context_parent = context._parent
        while context_parent:
            for key in context_parent:
                if context_parent[key] == context:
                    context_path.insert(0, key)
            context = context._parent
            context_parent = context_parent._parent
        return context_path

    def sort(self) -> JSON:
        pass

    def merge(self, secondary: JSON, path_separator: str = "/") -> Tuple[dict, List[str], List[str]]:
        # Merges the given secondary JSON object into a COPY of this JSON object; but does not overwrite
        # anything in this JSON object; anything that would otherwise overwrite is ignored. Returns a tuple
        # with (left-to-right) the (new) merged dictionary, a list of paths which were actually merged from
        # the secondary, and a list of paths which were not merged from the secondary, i.e. because they would
        # have overwritten that item in the copy of this JSON object; path delimiter is the given path_separator.
        return JSON._merge(self, secondary, path_separator=path_separator)

    def _initialize(self, parent: Optional[JSON] = None) -> None:
        if self._initialized is False:
            self._initialized = None
            if not parent:
                parent = self
            for key, value in parent.items():
                # value = super(JSON, parent).__getitem__(key)
                if isinstance(value, dict):
                    if not isinstance(value, JSON):
                        value = JSON(value, _initializing=True)
                    value._parent = parent
                    super(JSON, parent).__setitem__(key, value)
                    self._initialize(value)
            self._initialized = True

    def __setitem__(self, key: Any, value: Any) -> None:
        self._initialize()
        if isinstance(value, dict):
            if id(value._parent) != id(self):
                if isinstance(value, JSON):
                    copied_value = deepcopy(value)
                    value = copied_value
                else:
                    value = JSON(value)
                value._parent = self
        else:
            value = self._write_value(value)
        super().__setitem__(key, value)

    def __getitem__(self, key: Any) -> Any:
        self._initialize()
        value = super().__getitem__(key)
        if not isinstance(value, dict):
            value = self._read_value(value)
        return value

    def __delitem__(self, key: Any) -> None:
        self._initialize()
        super().__delitem__(key)

    def __iter__(self) -> Iterator[Any]:
        self._initialize()
        return super().__iter__()

    def __deepcopy__(self, memo) -> JSON:
        return JSON(deepcopy(dict(self), memo))

    @staticmethod
    def _merge(primary: JSON, secondary: JSON, path_separator: str = "/") -> Tuple[dict, List[str], List[str]]:
        # Merges the given secondary JSON object into a COPY of the given primary JSON object; but does not
        # overwrite anything in the primary JSON object; anything that would otherwise overwrite is ignored.
        # Returns a tuple with (in left-right order) the (new) merged dictionary, list of paths which were
        # actually merged from the secondary, and a list of paths which were not merged from the secondary,
        # i.e. because they would have overwritten in the primary; path delimiter is the given path_separator.
        if not (isinstance(primary, dict) or isinstance(secondary, dict)):
            return None, None, None
        merged = deepcopy(primary) ; merged_paths = [] ; unmerged_paths = []  # noqa
        def merge(primary: dict, secondary: dict, path: str = "") -> None:  # noqa
            nonlocal unmerged_paths, path_separator
            for key, value in secondary.items():
                key_path = f"{path}{path_separator}{key}" if path else key
                if key not in primary:
                    primary[key] = secondary[key]
                    merged_paths.append(key_path)
                elif isinstance(primary[key], dict) and isinstance(secondary[key], dict):
                    merge(primary[key], secondary[key], path=key_path)
                else:
                    unmerged_paths.append(key_path)
        merge(merged, secondary)
        return merged, merged_paths, unmerged_paths



from __future__ import annotations
import os
from typing import Any, Callable, List, Optional, Tuple, Union
from hms_utils.dictionary_parented import DictionaryParented as JSON
from hms_utils.type_utils import is_primitive_type, primitive_type
from hms_utils.config.config_basic import ConfigBasic


class ConfigWithSecrets(ConfigBasic):

    _SECRET_OBFUSCATED_VALUE = "********"
    _SECRET_VALUE_START = "@@@@@@@__mark_secret_start__["
    _SECRET_VALUE_END = "]__mark_secret_end__@@@@@@@"
    _SECRET_VALUE_START_LENGTH = len(_SECRET_VALUE_START)
    _SECRET_VALUE_END_LENGTH = len(_SECRET_VALUE_END)
    _TYPE_NAME_STR = type("").__name__
    _TYPE_NAME_INT = type(1).__name__
    _TYPE_NAME_FLOAT = type(1.0).__name__
    _TYPE_NAME_BOOL = type(True).__name__
    _TYPE_NAME_SEPARATOR = ":"
    _TYPE_NAME_LENGTH_STR = len(_TYPE_NAME_STR)
    _TYPE_NAME_LENGTH_INT = len(_TYPE_NAME_INT)
    _TYPE_NAME_LENGTH_FLOAT = len(_TYPE_NAME_FLOAT)
    _TYPE_NAME_LENGTH_BOOL = len(_TYPE_NAME_BOOL)
    _TYPE_NAME_LENGTH_SEPARATOR = len(_TYPE_NAME_SEPARATOR)

    def __init__(self, config: Union[dict, str],
                 name: Optional[str] = None,
                 path_separator: Optional[str] = None,
                 custom_macro_lookup: Optional[Callable] = None,
                 raise_exception: bool = False,
                 secrets: bool = False,
                 obfuscated_value: Optional[str] = None, **kwargs) -> None:
        self._secrets = (secrets is True) or (isinstance(config, str) and ("secret" in os.path.basename(config)))
        self._obfuscated_value = (obfuscated_value
                                  if isinstance(obfuscated_value, str) and obfuscated_value
                                  else ConfigWithSecrets._SECRET_OBFUSCATED_VALUE)
        super().__init__(config=config,
                         name=name,
                         path_separator=path_separator,
                         custom_macro_lookup=custom_macro_lookup,
                         raise_exception=raise_exception, **kwargs)

    def _create_json(self, data: dict) -> JSON:
        if not self._secrets:
            return super()._create_json(data)
        return JSON(data, rvalue=self._secrets_encoded if self._secrets else None)

    def data(self, show: Optional[bool] = False) -> JSON:
        if self._secrets:
            if show is True:
                return JSON(self.json, rvalue=self._secrets_plaintext).sorted()
            elif show is False:
                return JSON(self.json, rvalue=self._secrets_obfuscated).sorted()
        return super().data()

    @property
    def secrets(self) -> bool:
        return self._secrets

    def lookup(self, path: str,
               context: Optional[JSON] = None,
               noexpand: bool = False,
               inherit_simple: bool = False,
               inherit_none: bool = False,
               show: Optional[bool] = False) -> Optional[Union[Any, JSON]]:
        value = super().lookup(path=path, context=context,
                               noexpand=noexpand, inherit_simple=inherit_simple, inherit_none=inherit_none, show=show)
        if self._secrets:
            if value is None:
                return value
            elif show is True:
                if isinstance(value, JSON):
                    return value.duplicate(rvalue=self._secrets_plaintext)
                else:
                    return self._secrets_plaintext(value)
            elif show is False:
                if isinstance(value, JSON):
                    return value.duplicate(rvalue=self._secrets_obfuscated)
                else:
                    return self._secrets_obfuscated(value)
        return value

    def merge(self, data: Union[Union[dict, ConfigBasic],
                                List[Union[dict, ConfigBasic]]]) -> Tuple[List[str], List[str]]:
        if isinstance(data, ConfigWithSecrets):
            self._secrets = data._secrets
        elif isinstance(data, list):
            for item in data:
                if isinstance(item, ConfigWithSecrets):
                    self._secrets = self._secrets or item._secrets
        return super().merge(data)

    # All of this marking of secrets stuff is just so that when obtaining values (for lookup/print/dump/display),
    # any strings which came from a "secret" configuration can be obfuscated by default, or shown if desired;
    # we do go to the trouble of not marking macros within secret config values as secret.

    def _secrets_encoded(self, value: primitive_type, value_type: Optional[str] = None) -> str:
        if not is_primitive_type(value):
            return ""
        if (actual_value_type := type(value).__name__) != ConfigWithSecrets._TYPE_NAME_STR:
            return (f"{ConfigWithSecrets._SECRET_VALUE_START}{actual_value_type}"
                    f"{ConfigWithSecrets._TYPE_NAME_SEPARATOR}{value}{ConfigWithSecrets._SECRET_VALUE_END}")
        if not (isinstance(value_type, str) and value_type):
            value_type = ConfigWithSecrets._TYPE_NAME_STR
        secrets_encoded = ""
        start = 0
        while True:
            if not (match := ConfigBasic._MACRO_PATTERN.search(value[start:])):
                if secret_part := value[start:]:
                    secrets_encoded += (
                        f"{ConfigWithSecrets._SECRET_VALUE_START}{value_type}"
                        f"{ConfigWithSecrets._TYPE_NAME_SEPARATOR}{secret_part}{ConfigWithSecrets._SECRET_VALUE_END}")
                break
            match_start = match.start()
            match_end = match.end()
            macro_part = value[start + match_start:start + match_end]
            if secret_part := value[start:start + match_start]:
                secrets_encoded += (
                    f"{ConfigWithSecrets._SECRET_VALUE_START}{value_type}"
                    f"{ConfigWithSecrets._TYPE_NAME_SEPARATOR}{secret_part}{ConfigWithSecrets._SECRET_VALUE_END}")
            secrets_encoded += macro_part
            start += match_end
        return secrets_encoded

    def _secrets_plaintext(self, value: Any, plaintext_value: Optional[Callable] = None) -> Any:
        if not callable(plaintext_value):
            plaintext_value = None
        if (not isinstance(value, str)) or (not value):
            if isinstance(value, dict):
                json = isinstance(value, JSON)
                value = {k: self._secrets_plaintext(v, plaintext_value=plaintext_value) for k, v in value.items()}
                return JSON(value) if json else value
            elif isinstance(value, list):
                return [self._secrets_plaintext(e, plaintext_value=plaintext_value) for e in value]
            return value
        secret_value_typed = None
        while True:
            if (start := value.find(ConfigWithSecrets._SECRET_VALUE_START)) < 0:
                break
            if (end := value.find(ConfigWithSecrets._SECRET_VALUE_END)) < start:
                break
            secret_value = value[start + ConfigWithSecrets._SECRET_VALUE_START_LENGTH:end]
            secret_value, secret_value_typed = self._secrets_plaintext_value(secret_value)
            if plaintext_value:
                secret_value = plaintext_value(secret_value)
            value = value[0:start] + secret_value + value[end + ConfigWithSecrets._SECRET_VALUE_END_LENGTH:]
        if (secret_value_typed is not None) and (str(secret_value_typed) == secret_value):
            return secret_value_typed
        return value

    def _secrets_plaintext_value(self, value: str) -> Tuple[primitive_type, Any]:
        from hms_utils.config.config_with_aws_macros import ConfigWithAwsMacros  # here to avoid circular imports
        value_typed = None
        if value.startswith(f"{ConfigWithSecrets._TYPE_NAME_STR}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"):
            value = value[ConfigWithSecrets._TYPE_NAME_LENGTH_STR + ConfigWithSecrets._TYPE_NAME_LENGTH_SEPARATOR:]
        elif value.startswith(f"{ConfigWithSecrets._TYPE_NAME_INT}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"):
            value = value[ConfigWithSecrets._TYPE_NAME_LENGTH_INT + ConfigWithSecrets._TYPE_NAME_LENGTH_SEPARATOR:]
            value_typed = int(value)
        elif value.startswith(f"{ConfigWithSecrets._TYPE_NAME_FLOAT}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"):
            value = value[ConfigWithSecrets._TYPE_NAME_LENGTH_FLOAT + ConfigWithSecrets._TYPE_NAME_LENGTH_SEPARATOR:]
            value_typed = float(value)
        elif value.startswith(f"{ConfigWithSecrets._TYPE_NAME_BOOL}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"):
            value = value[ConfigWithSecrets._TYPE_NAME_LENGTH_BOOL + ConfigWithSecrets._TYPE_NAME_LENGTH_SEPARATOR:]
            value_typed = True if (value.lower() == "true") else False
        elif isinstance(self, ConfigWithAwsMacros):
            if (aws_value := ConfigWithAwsMacros._secrets_plaintext_value(self, value)) is not None:
                value = aws_value
        return value, value_typed

    def _secrets_obfuscated(self, value: Any, obfuscated_value: Optional[Union[str, Callable]] = None) -> Any:
        if (not isinstance(value, str)) or (not value):
            if isinstance(value, dict):
                json = isinstance(value, JSON)
                value = {k: self._secrets_obfuscated(v, obfuscated_value=obfuscated_value) for k, v in value.items()}
                return JSON(value) if json else value
            elif isinstance(value, list):
                return [self._secrets_obfuscated(e, obfuscated_value=obfuscated_value) for e in value]
            return value
        if not (callable(obfuscated_value) or ((isinstance(obfuscated_value, str)) and obfuscated_value)):
            obfuscated_value = self._obfuscated_value
        while True:
            if (start := value.find(ConfigWithSecrets._SECRET_VALUE_START)) < 0:
                break
            if (end := value.find(ConfigWithSecrets._SECRET_VALUE_END)) < start:
                break
            if callable(obfuscated_value):
                obfuscated = (obfuscated_value(self._obfuscated_value)
                              if end > start + ConfigWithSecrets._SECRET_VALUE_START_LENGTH else "")
            else:
                obfuscated = obfuscated_value if end > start + ConfigWithSecrets._SECRET_VALUE_START_LENGTH else ""
            value = value[0:start] + obfuscated + value[end + ConfigWithSecrets._SECRET_VALUE_END_LENGTH:]
        return value

    def _contains_secret_values(self, value: Any) -> bool:
        if not isinstance(value, str):
            return False
        if (start := value.find(ConfigWithSecrets._SECRET_VALUE_START)) < 0:
            return False
        return value.find(ConfigWithSecrets._SECRET_VALUE_END) > start



from importlib.metadata import version as get_package_version


def get_version(package_name: str = "hms-utils") -> str:
    try:
        return get_package_version(package_name)
    except Exception:
        return ""



import os
from typing import List, Optional


def unpack_path(path: str,
                path_separator: Optional[str] = None,
                path_current: Optional[str] = None,
                path_parent: Optional[str] = None,
                noroot: bool = False) -> List[str]:
    if not (isinstance(path_separator, str) and (path_separator := path_separator.strip())):
        path_separator = os.sep
    if not (isinstance(path_current, str) and (path_current := path_current.strip())):
        path_current = "."
    if not (isinstance(path_parent, str) and (path_parent := path_parent.strip())):
        path_parent = "."
    path_components = []
    if isinstance(path, str) and (path := path.strip()):
        if path.startswith(path_separator):
            path = path[len(path_separator):]
            path_components.append(path_separator)
        for path_component in path.split(path_separator):
            if ((path_component := path_component.strip()) and (path_component != path_current)):
                if path_component == path_parent:
                    if (len(path_components) > 0) and (path_components != [path_separator]):
                        path_components = path_components[:-1]
                    continue
                path_components.append(path_component)
    if (noroot is True) and path_components and path_components[0] == path_separator:
        path_components = path_components[1:]
    return path_components


def repack_path(path_components: List[str], path_separator: Optional[str] = None, path_rooted: bool = False) -> str:
    if not (isinstance(path_separator, str) and (path_separator := path_separator.strip())):
        path_separator = os.sep
    if not (isinstance(path_components, list) and path_components):
        path_components = []
    if path_components[0] == path_separator:
        path_rooted = True
        path_components = path_components[1:]
    return (path_separator if path_rooted else "") + path_separator.join(path_components)


def basename_path(path: str, path_separator: Optional[str] = None) -> str:
    if not (isinstance(path, str) and (path := path.strip())):
        return ""
    if not (isinstance(path_separator, str) and (path_separator := path_separator.strip())):
        path_separator = os.sep
    if (index := path.rfind(path_separator)) > 0:
        return path[index + 1:]
    return path


def is_current_or_parent_relative_path(path: str) -> bool:
    return isinstance(path, str) and path.startswith(f".{os.sep}") or path.startswith(f"..{os.sep}")



from prettytable import PrettyTable
from typing import Callable, List, Optional
from hms_utils.chars import chars
from hms_utils.type_utils import is_uuid


def print_principals(principals: List[str], message: Optional[str] = None,
                     value_callback: Optional[Callable] = None,
                     value_header: Optional[str] = None,
                     return_value: bool = False,
                     nosort: bool = False):

    def uuid_specific_principal(principal):  # noqa
        principal_uuid = ""
        if principal.startswith("role."):
            principal = principal.replace("role.", "role/")
        if (index := principal.find(".")) > 0:
            if is_uuid(value := principal[index + 1:]):
                principal_uuid = value
                principal = principal[0:index]
        if principal.startswith("role/"):
            principal = principal.replace("role/", "role.")
        return principal, principal_uuid

    rows = [] ; header = None  # noqa
    for principal in principals:
        principal, principal_uuid = uuid_specific_principal(principal)
        if principal_uuid:
            if (principal == "userid") and (not header):
                header = [principal_uuid, "ROLE UUID"]
                if value_callback:
                    header.append(value_header or "")
            else:
                rows.append([principal, principal_uuid])
                if value_callback:
                    rows[len(rows) - 1].append("")
        else:
            rows.append([principal, ""])
            if value_callback:
                rows[len(rows) - 1].append("")

    table = PrettyTable()
    if header:
        table.field_names = header
        if value_callback and (role := value_callback("userid", header[0])):
            rows.append(["userid", header[0], role])
    else:
        table.header = False
    for row in rows:
        if value_callback and (role := value_callback(row[0], row[1])):
            row[2] = role
    if nosort is not True:
        rows.sort(key=lambda row: f"{row[0]}_{row[1]}")
    for row in rows:
        table.add_row([item or chars.null for item in row])
    table.align = "l"
    output = f"{message}\n" if message else ""
    output += str(table)
    if return_value:
        return output
    print(output)


def print_roles(roles: dict, message: Optional[str] = None, nosort: bool = False):
    def role_value(principal, uuid):
        nonlocal roles
        return roles.get(f"{principal}.{uuid}") if uuid else roles.get(principal)
    print_principals(list(roles.keys()), message=message,
                     value_callback=role_value, value_header="ROLE VALUE", nosort=nosort)


def print_acls(acls: List[tuple], message: Optional[str] = None, nosort: bool = False):
    rows = []
    for acl_item in acls:
        ace_action, ace_principal, ace_permissions = acl_item
        rows.append([ace_principal,
                     (f"{ace_action} {chars.dot}"
                      f" {'/'.join(ace_permissions) if isinstance(ace_permissions, list) else str(ace_permissions)}")])
    table = PrettyTable()
    table.header = False
    if nosort is not True:
        rows.sort(key=lambda row: row[0])
    for row in rows:
        table.add_row(row)
    table.align = "l"
    if message:
        print(message)
    print(table)


def print_acls_and_principals(acls: List[tuple], principals: List[str],
                              message: Optional[str] = None, permission: Optional[str] = None):
    def find_matching_acl():  # noqa
        # FYI: See pyramid/authorization.py ...
        nonlocal acls, principals, permission
        for acl_item in acls:
            acl_action, acl_principal, acl_permissions = acl_item
            if acl_principal in principals:
                if permission in acl_permissions:
                    return acl_item
        return None
    matching_acl = find_matching_acl()
    def acl_value(principal, uuid):  # noqa
        nonlocal acls, principals, permission
        if uuid:
            principal = f"{principal}.{uuid}"
        for acl_item in acls:
            acl_action, acl_principal, acl_permissions = acl_item
            if acl_principal == principal:
                value = (f"{acl_action} {chars.dot}"
                         f" {'/'.join(acl_permissions) if isinstance(acl_permissions, list) else str(acl_permissions)}")
                if acl_item == matching_acl:
                    value += f" {chars.larrow}{chars.larrow}{chars.larrow}"
                elif permission in acl_permissions:
                    value += f" {chars.larrow_hollow}"
                return value
        return ""
    output = print_principals(principals, value_callback=acl_value, value_header="PERMISSION", return_value=True)
    acls_not_in_principals = []
    for acl_item in acls:
        if acl_item[1] not in principals:
            acls_not_in_principals.append(acl_item)
    if acls_not_in_principals:
        if (index := output.rfind("\n")) > 0:
            separator = output[index + 1:].replace("+", "-")
            separator = "+" + ((len(separator) - 2) * "-") + "+"
        output += "\n| ACLs NOT IN PRINCIPALS ...\n" + separator
        for acl_item in acls_not_in_principals:
            acl_item_output = (f"| {acl_item[1]} {chars.dot} {acl_item[0]} {chars.dot}"
                               f" {'/'.join(acl_item[2]) if isinstance(acl_item[2], list) else str(acl_item[2])}")
            output += f"\n{acl_item_output}"
            output += ((len(separator) - len(acl_item_output) - 1) * " ") + "|"
        output += "\n" + separator
    if message:
        print(message)
    print(output)


if False:

    import json
    principals = ['group.submitter', 'role.consortium_member_create', 'accesskey.7JFE33AN', 'role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff', 'system.Authenticated', 'group.admin', 'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff', 'system.Everyone', 'role.consortium_member_rw', 'userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b']  # noqa
    print(json.dumps(principals, indent=4))
    print_principals(principals)


if False:

    acls = [('Allow', 'group.submitter', ['edit']), ('Allow', 'role.submission_center_member_rw', ['view']), ('Allow', 'group.admin', ['view', 'edit']), ('Allow', 'group.read-only-admin', ['view']), ('Allow', 'remoteuser.INDEXER', ['view']), ('Allow', 'remoteuser.EMBED', ['view']), ('Deny', 'system.Everyone', ['view', 'edit'])]  # noqa

    principals = {'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff', 'system.Everyone', 'group.admin', 'accesskey.7JFE33AN', 'role.owner', 'role.submission_center_member_create', 'userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68', 'role.submission_center_member_rw', 'group.submitter', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b', 'role.consortium_member_create', 'role.consortium_member_rw', 'system.Authenticated', 'role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff'}  # noqa

    print_acls(acls)
    print_principals(principals)
    print_acls_and_principals(acls, principals, permission="view")


if False:
    acl = [('Allow', 'group.submitter', ['edit']), ('Allow', 'role.submission_center_member_rw', ['view']), ('Allow', 'group.admin', ['view', 'edit']), ('Allow', 'group.read-only-admin', ['view']), ('Allow', 'remoteuser.INDEXER', ['view']), ('Allow', 'remoteuser.EMBED', ['view']), ('Deny', 'system.Everyone', ['view', 'edit'])]  # noqa
    principals = {'accesskey.JMGZMR42', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b', 'role.submission_center_member_rw.055cfde3-f82c-4d32-851d-b8061c7e2b95', 'system.Everyone', 'role.consortium_member_create', 'role.consortium_member_rw', 'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff', 'system.Authenticated', 'role.submission_center_member_create', 'group.submitter', 'userid.0d565156-00db-4948-82fe-021d151a5daf'}  # noqa
    print_acls_and_principals(acl, principals)

if False:

    acls = [('Allow', 'group.admin', ['view', 'edit']), ('Allow', 'group.read-only-admin', ['view']), ('Allow', 'remoteuser.INDEXER', ['view']), ('Allow', 'remoteuser.EMBED', ['view']), ('Deny', 'system.Everyone', ['view', 'edit'])]  # noqa
    print_acls(acls)
    principals = ['system.Everyone', 'group.admin', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b', 'role.consortium_member_rw', 'system.Authenticated', 'userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68', 'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff', 'role.consortium_member_create', 'accesskey.Z4WJPLBI', 'group.submitter', 'role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff']  # noqa
    print_principals(principals)

    print_acls_and_principals(acls, principals, permission="view")

    roles = {'role.submission_center_member_rw.9626d82e-8110-4213-ac75-0a50adf890ff': 'role.submission_center_member_rw', 'submits_for.9626d82e-8110-4213-ac75-0a50adf890ff': 'role.submission_center_member_create', 'role.consortium_member_rw.358aed10-9b9d-4e26-ab84-4bd162da182b': 'role.consortium_member_rw', 'userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68': 'role.owner'}  # noqa
    print_roles(roles)



import json
from dcicutils.ff_utils import search_metadata
from dcicutils.portal_utils import Portal
from hms_utils.argv import ARGV


def main():

    argv = ARGV({
        ARGV.REQUIRED(str): ["--env", ARGV.DEFAULT, "smaht-local"],
        ARGV.OPTIONAL(bool): ["--verbose"]
    })

    portal = Portal(argv.env)
    status = 0

    query = "/search/?type=Item&consortia.display_title=No+value&limit=1000000"
    ignored_types = ["AccessKey", "TrackingItem", "Consortium", "SubmissionCenter"]
    total = 0
    for item in search_metadata(query, key=portal.key, is_generator=True):
        item_type = portal.get_schema_type(item)
        if item_type in ignored_types:
            continue
        item_uuid = item.get("uuid")
        print(f"{item_uuid}: {item_type}")
        print(json.dumps(item, indent=4))
        total += 1

    print(f"TOTAL: {total}")
    exit(status)


if __name__ == "__main__":
    main()
    pass



from __future__ import annotations
import os
import pyramid
import sys
import webtest
from typing import Callable, List, Optional, Union
from webob.multidict import MultiDict
from dcicutils.captured_output import captured_output
from dcicutils.portal_utils import Portal as PortalFromUtils
from dcicutils.ff_utils import delete_field, delete_metadata, purge_metadata
from dcicutils.common import APP_SMAHT, ORCHESTRATED_APPS
from hms_utils.chars import chars
from hms_utils.type_utils import is_uuid, to_non_empty_string_list


class Portal(PortalFromUtils):

    def get_metadata(self, object_id: str, raw: bool = False, database: bool = False,
                     limit: Optional[int] = None, offset: Optional[int] = None,
                     field: Optional[str] = None, deleted: bool = False,
                     raise_exception: bool = True) -> Optional[dict]:
        if isinstance(object_id, str):
            object_id = object_id.lstrip("/")
        return super().get_metadata(object_id, raw, database, limit, offset, field, deleted, raise_exception)

    def delete_metadata(self, object_id: str) -> Optional[dict]:
        if self.key:
            return delete_metadata(obj_id=object_id, key=self.key)
        return None

    def purge_metadata(self, object_id: str) -> Optional[dict]:
        if self.key:
            return purge_metadata(obj_id=object_id, key=self.key)
        return None

    def delete_metadata_property(self, item_path: str, property_name: str, raise_exception: bool = False) -> bool:
        try:
            delete_field(item_path, property_name, key=self.key)
            return True
        except Exception as e:
            if raise_exception is True:
                raise e
        return False

    def reindex_metadata(self, uuids: Union[List[str], str], raise_exception: bool = False) -> bool:
        if isinstance(uuids, str) and is_uuid(uuids):
            uuids = [uuids]
        elif isinstance(uuids, list):
            if not (uuids := [uuid for uuid in uuids if is_uuid(uuid)]):
                return False
        else:
            return False
        try:
            return self.post("/queue_indexing", json={"uuids": uuids}).status_code == 200
        except Exception as e:
            if raise_exception is True:
                raise e
            return False

    @staticmethod
    def get_item_type(item: dict) -> Optional[str]:
        if isinstance(item, dict):
            if isinstance(item_types := item.get("@type"), list):
                return item_types[0] if (item_types := to_non_empty_string_list(item_types)) else None
            return item_types if isinstance(item_types, str) and item_types else None
        return None

    @classmethod
    def create(cls,
               arg: Optional[str] = None,
               app: Optional[str] = None,
               raise_exception: bool = False,
               verbose: bool = False,
               debug: bool = False,
               noerror: bool = False,
               ping: bool = False,
               noping: bool = False,
               show: bool = False,
               printf: Optional[Callable] = None, **kwargs) -> Portal:

        env_environ_name = None
        if not (isinstance(arg, str) and arg):
            if isinstance(app, str) and app and (app in ORCHESTRATED_APPS):
                env_app_name = app
            else:
                env_app_name = APP_SMAHT
            env_environ_name = f"{env_app_name.upper()}_ENV"
            if not (arg := os.environ.get(env_environ_name)):
                env_environ_name = None

        if not callable(printf):
            printf = lambda *args, **kwargs: print(*args, **kwargs, file=sys.stderr)  # noqa

        if ping:
            verbose = True

        with captured_output(debug is not True):
            try:
                portal = cls(arg, app=app, raise_exception=raise_exception, **kwargs)
            except Exception as e:
                if raise_exception is True:
                    raise
                if (verbose is True) or (noerror is not True):
                    printf(f"ERROR: {str(e)}")
                return None

        if verbose is True:
            if portal.env:
                if env_environ_name:
                    printf(f"Portal environment: {portal.env}"
                           f" {chars.dot} from environment variable: {env_environ_name}")
                else:
                    printf(f"Portal environment: {portal.env}")
            if portal.keys_file:
                printf(f"Portal keys file: {portal.keys_file}")
            if portal.key_id:
                if show is True:
                    printf(f"Portal key: {portal.key_id} {chars.dot} {portal.secret}")
                else:
                    printf(f"Portal key prefix: {portal.key_id[0:2]}******")
            if portal.ini_file:
                printf(f"Portal ini file: {portal.ini_file}")
            if portal.server:
                printf(f"Portal server: {portal.server}")
        if noping is not True:
            status = 0
            if portal.ping():
                if verbose:
                    version = None
                    bluegreen = None
                    try:
                        # SMaHT only just for BTW/FYI ...
                        health = portal.get_health().json()
                        version = health.get("project_version")
                        if (beanstalk_env := health.get("beanstalk_env")) == "smaht-production-green":
                            bluegreen = "green"
                        elif beanstalk_env == "smaht-production-blue":
                            bluegreen = "blue"
                    except Exception:
                        pass
                    printf(f"Portal connectivity: OK {chars.dot}{f' {version}' if version else ''}"
                           f"{f' {chars.dot} {bluegreen}' if bluegreen else ''} {chars.check}")
            else:
                printf(f"Portal connectivity: {portal.server} ({portal.env}) is unreachable {chars.xmark}")
                return None
            if ping is True:
                exit(status)

        return portal


class PyramidRequestForTesting(pyramid.request.Request):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._params = None
        self.vapp = None
    @property  # noqa
    def params(self) -> MultiDict:
        if self._params is not None:
            return self._params
        return super().params
    @params.setter  # noqa
    def params(self, value: MultiDict) -> None:
        if not isinstance(value, MultiDict):
            raise Exception("Must set pyramid.request.Request.params to MultiDict.")
        self._params = value


def create_pyramid_request_for_testing(
        portal_or_vapp_or_request: Union[PortalFromUtils, webtest.app.TestApp],
        args: Optional[Union[str, dict]] = None) -> Optional[pyramid.request.Request]:
    if not isinstance(vapp := portal_or_vapp_or_request, webtest.app.TestApp):
        if not (isinstance(portal_or_vapp_or_request, PortalFromUtils) and
                isinstance(vapp := portal_or_vapp_or_request.vapp, webtest.app.TestApp)):
            if not (isinstance(portal_or_vapp_or_request, PyramidRequestForTesting) and
                    isinstance(vapp := portal_or_vapp_or_request.vapp, webtest.app.TestApp)):
                return None
    if not isinstance(router := vapp.app, pyramid.router.Router):
        return None
    if not isinstance(registry := router.registry, pyramid.registry.Registry):
        return None
    if not isinstance(response := vapp.get("/"), webtest.response.TestResponse):
        return None
    if not isinstance(request := response.request, webtest.app.TestRequest):
        return None
    if not isinstance(environ := request.environ, dict):
        return None
    request = PyramidRequestForTesting(environ)
    request.registry = registry
    request.vapp = vapp
    if isinstance(args, dict):
        params = []
        for key in args:
            if isinstance(value := args[key], (list, tuple)):
                for item in value:
                    params.append((key, item))
            else:
                params.append((key, value))
        request.params = MultiDict(params)
    elif isinstance(args, MultiDict):
        request.params = args
    return request


def portal_custom_search(portal_or_vapp_or_request: Union[PortalFromUtils,
                                                          webtest.app.TestApp, pyramid.request.Request],
                         query: str,
                         method: Optional[str] = None,
                         aggregations: Optional[dict] = None) -> Optional[dict]:
    try:
        from snovault.search.search import search as snovault_search
        from snovault.search.search_utils import make_search_subreq as snovault_make_search_subreq
    except Exception:
        print("ERROR: Cannot execute this code outside of snovault context!", file=sys.stderr, flush=True)
        exit(1)
    if not (isinstance(method, str) and (method := method.strip())):
        method = "GET"
    if not isinstance(request := portal_or_vapp_or_request, pyramid.request.Request):
        if not (request := create_pyramid_request_for_testing(portal_or_vapp_or_request)):
            return None
    if not isinstance(query, str):
        query = "/"
    if not isinstance(aggregations, dict):
        aggregations = None
    request = snovault_make_search_subreq(request, path=query, method=method)
    return snovault_search(None, request, custom_aggregations=aggregations)



from datetime import date, datetime, timedelta, timezone
from typing import Optional, Union
from dcicutils.datetime_utils import parse_datetime_string as dcicutils_parse_datetime_string


def convert_uptime_to_datetime(uptime: str, relative_to: datetime = None) -> Optional[datetime]:
    """
    Converts the given duration string which (happens to be) from the Portal health endpoint
    into its equivalent datetime. Format of the given upatime looks something like this:

      1 week, 2 days, 3 hours, 4 minutes, 5.67 seconds

    If the given uptime is not parsable then returns None, otherwise returns the datetime corresponding
    to this uptime, relative to now, by default, or to the relative_to datetime argument if given.
    We THINK it's right to interpret the given uptime relative to UTC (TODO).
    """
    def normalize_spaces(s: str) -> str:
        return " ".join(s.split())

    if not uptime:
        return None
    try:
        minutes_per_hour = 60
        minutes_per_day = minutes_per_hour * 24
        minutes_per_week = minutes_per_day * 7
        minutes = 0
        seconds = 0
        uptime = normalize_spaces(uptime)
        for item in uptime.split(","):
            item = item.strip()
            if item:
                item = item.split(" ")
                if len(item) == 2:
                    unit = item[1].lower()
                    value = float(item[0])
                    if unit.startswith("week"):
                        minutes += minutes_per_week * value
                    elif unit.startswith("day"):
                        minutes += minutes_per_day * value
                    elif unit.startswith("hour"):
                        minutes += minutes_per_hour * value
                    elif unit.startswith("minute"):
                        minutes += value
                    elif unit.startswith("second"):
                        seconds += value
        t = relative_to if relative_to else datetime.now(timezone.utc)
        return t + timedelta(minutes=-minutes, seconds=-seconds)
    except Exception:
        pass
    return None


def format_duration(seconds: Union[int, float, timedelta], verbose: bool = False) -> str:
    if verbose is True:
        return format_duration_verbose(seconds)
    if isinstance(seconds, timedelta):
        seconds = seconds.total_seconds()
    days = (duration := timedelta(seconds=seconds)).days
    hours, remainder = divmod(duration.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if days > 0:
        return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def format_duration_verbose(seconds: Union[int, float, timedelta]) -> str:
    if isinstance(seconds, timedelta):
        seconds = seconds.total_seconds()
    seconds_actual = seconds
    seconds = round(max(seconds, 0))
    durations = [("year", 31536000), ("day", 86400), ("hour", 3600), ("minute", 60), ("second", 1)]
    parts = []
    for name, duration in durations:
        if (seconds == 0) or (seconds >= duration):
            count = seconds // duration
            seconds %= duration
            if count != 1:
                name += "s"
            parts.append(f"{count} {name}")
    if len(parts) == 0:
        return f"{seconds_actual:.1f} seconds"
    elif len(parts) == 1:
        return f"{seconds_actual:.1f} seconds"
    return " ".join(parts[:-1]) + " " + parts[-1]


def parse_datetime_string(value: str) -> Optional[datetime]:
    if isinstance(value, str) and (len(value) == 8) and value.isdigit():
        # Very special case to accept for example "20241206" to mean "2024-12-06".
        value = f"{value[0:4]}-{value[4:6]}-{value[6:8]}"
    elif isinstance(value, datetime):
        return value
    elif isinstance(value, date):
        return datetime.combine(value, datetime.min.time())
    return dcicutils_parse_datetime_string(value)



from __future__ import annotations
from functools import lru_cache
import glob
import io
import json
import os
import sys
from typing import Any, Callable, List, Optional, Union
from dcicutils.portal_utils import Portal as PortalFromUtils
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.dictionary_utils import sort_dictionary
from hms_utils.portal.portal_utils import Portal as Portal
from hms_utils.threading_utils import run_concurrently

_ITEM_UUID_PROPERTY_NAME = "uuid"


def main():

    argv = ARGV({
        ARGV.REQUIRED(str): ["directory"],
        ARGV.OPTIONAL(str): ["--env"],
        ARGV.OPTIONAL(str): ["--app"],
        ARGV.OPTIONAL(int, 32): ["--threads"],
        ARGV.OPTIONAL(bool): ["--ping"],
        ARGV.OPTIONAL(bool, False): ["--verbose"],
        ARGV.OPTIONAL(bool, False): ["--debug"],
    })

    if not os.path.isdir(directory := os.path.expanduser(argv.directory)):
        print(f"Cannot find directory: {directory}")
        return 1

    nitems_total = 0
    nfiles_total = 0

    for file in glob.glob(os.path.join(directory, "*.json")):
        if not Portal.schema_name(file):
            print(f"WARNING: File name does not corresond to known item type name: {file}")
            continue
        try:
            with io.open(file, "r") as f:
                if not isinstance(items := json.load(f), list):
                    print(f"WARNING: File does not contain JSON list: {file}")
                    continue
                nfiles_total += 1
                nitems = len(items)
                nitems_total += nitems
        except Exception:
            print(f"WARNING: Exception loading JSON from file: {file}")
            continue

    if argv.verbose:
        print(f"Checking JSON files in directory for Portal conflicts: {directory}")

    if not (portal := Portal.create(argv.env, app=argv.app, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)):
        print(f"Cannot access Portal: {argv.env}")
        return 1

    if argv.verbose:
        print(f"Total files to check: {nfiles_total}")
        print(f"Total items to check: items: {nitems_total}")
        if argv.threads > 1:
            print(f"Checking concurrency: {argv.threads} threads")

    for file in glob.glob(os.path.join(directory, "*.json")):
        check_file_for_conflicts(portal, file, threads=argv.threads, verbose=argv.verbose, debug=argv.debug)


def check_file_for_conflicts(portal: Portal, file: str,
                             threads: int = 0, verbose: bool = False, debug: bool = False) -> None:

    if not (item_type := Portal.schema_name(file)):
        print(f"WARNING: File name does not corresond to known item type name: {file}")
        return
    try:
        with io.open(file, "r") as f:
            if not isinstance(items := json.load(f), list):
                print(f"WARNING: File does not contain JSON list: {file}")
                return
    except Exception:
        print(f"WARNING: Exception loading JSON from file: {file}")
        return

    def check_item_for_conflicts_function(item: dict) -> None:  # noqa
        nonlocal portal, file, item_type
        check_item_for_conflicts(portal, item=item, item_type=item_type, item_source=file,
                                 report=True, debug=debug)
    if (verbose is True) or (debug is True):
        print(f"Checking file for conflicts: {file}")
    if isinstance(threads, int) and (threads > 0):
        check_item_for_conflicts_functions = []
        for item in items:
            check_item_for_conflicts_functions.append(lambda item=item: check_item_for_conflicts_function(item))
        run_concurrently(check_item_for_conflicts_functions, nthreads=threads)
    else:
        for item in items:
            check_item_for_conflicts(portal, item=item, item_type=item_type, item_source=file,
                                     report=True, debug=debug)


def check_item_for_conflicts(portal: Portal, item: dict, item_type: Optional[str] = None,
                             item_source: Optional[str] = None,
                             report: bool = False, printf: Optional[Callable] = None,
                             debug: bool = False) -> Union[List[dict], bool]:

    def identifying_values_are_equal(item_identifying_value: Any, existing_item_identifying_value: Any) -> bool:
        if isinstance(item_identifying_value, list):
            item_identifying_value = sorted(item_identifying_value)
        if isinstance(existing_item_identifying_value, list):
            existing_item_identifying_value = sorted(existing_item_identifying_value)
        return item_identifying_value == existing_item_identifying_value

    def reorder_item_properties(item: dict) -> None:
        if isinstance(item, dict):
            item = sort_dictionary(item)
            if (uuid := item.get(_ITEM_UUID_PROPERTY_NAME)) is not None:
                del item[_ITEM_UUID_PROPERTY_NAME]
                item = {_ITEM_UUID_PROPERTY_NAME: uuid, **item}
        return item

    if not isinstance(portal, (Portal, PortalFromUtils)):
        return False if report is True else []

    if not (identifying_properties := portal.get_identifying_property_names(item_type)):
        return False if report is True else []

    if _ITEM_UUID_PROPERTY_NAME in identifying_properties:
        identifying_properties.remove(_ITEM_UUID_PROPERTY_NAME)
        if not identifying_properties:
            return False if report is True else []

    if not callable(printf):
        printf = print

    conflicts = []

    item_uuid = item.get(_ITEM_UUID_PROPERTY_NAME)
    if item_uuid == "11f94a17-51ed-4a0f-93b1-1cac2fd2844f":
        pass

    existing_items_found = 0

    item_conflicts = []
    for identifying_property in identifying_properties:
        if (identifying_values := item.get(identifying_property)) is not None:
            if not isinstance(identifying_values, list):
                identifying_values = [identifying_values]
            for identifying_value in identifying_values:
                if (existing_item := get_portal_item_metadata(portal, item_type,
                                                              identifying_property, identifying_value)) is not None:
                    existing_items_found += 1
                    if (existing_item_uuid := existing_item.get(_ITEM_UUID_PROPERTY_NAME)) != item_uuid:
                        item_conflicts.append({
                            "identifying_property": identifying_property,
                            "identifying_value": identifying_value,
                            "item_uuid": item_uuid,
                            "existing_item_uuid": existing_item_uuid
                        })
    if item_conflicts:
        conflicts.append({
            "conflict": {
                "type": item_type,
                "identifying_properties": identifying_properties,
                "conflicts": item_conflicts,
                "item_source": item_source,
                "item": reorder_item_properties(item),
                "existing_source": portal.env,
                "existing_item": reorder_item_properties(existing_item)
            }
        })

    if (existing_item := get_portal_item_metadata(portal, item_type, _ITEM_UUID_PROPERTY_NAME, item_uuid)) is not None:
        item_conflicts = []
        existing_items_found += 1
        for identifying_property in identifying_properties:
            if (item_identifying_value := item.get(identifying_property)) is None:
                continue
            if (existing_item_identifying_value := existing_item.get(identifying_property)) is None:
                continue
            if not identifying_values_are_equal(item_identifying_value, existing_item_identifying_value):
                item_conflicts.append({
                    "identifying_property": identifying_property,
                    "identifying_value": item_identifying_value,
                    "existing_identifying_value": existing_item_identifying_value
                })
        if item_conflicts:
            conflicts.append({
                "conflict": {
                    "type": item_type,
                    "uuid": item_uuid,
                    "identifying_properties": identifying_properties,
                    "conflicts": item_conflicts,
                    "item_source": item_source,
                    "item": reorder_item_properties(item),
                    "existing_source": portal.env,
                    "existing_item": reorder_item_properties(existing_item)
                }
            })

    if debug is True:
        message = f"Checking item for conflicts: /{item_type}/{item_uuid} ... "
        if conflicts:
            message += "conflicts found:"
            message += "\n" + json.dumps(conflicts, indent=4)
        else:
            message += f"OK {chars.dot} found: {existing_items_found}"
        printf(message)
    elif (report is True) and conflicts:
        message = f"Conflicts found in item: /{item_type}/{item_uuid}"
        message += "\n" + json.dumps(conflicts, indent=4)
        printf(message)

    return conflicts if (report is not True) else (len(conflicts) > 0)


@lru_cache(maxsize=1024)
def get_portal_item_metadata(portal: Portal, item_type: str,
                             identifying_property: str, identifying_value: Any) -> Optional[dict]:
    return portal.get_metadata(f"/{item_type}/{identifying_value}", raw=True, raise_exception=False)


if __name__ == "__main__":
    status = main()
    sys.exit(status if isinstance(status, int) else 0)
